@@ -56,8 +56,9 @@ from .core import compute_up, compute, base
 from ..expr import (
     Projection, Selection, Field, Broadcast, Expr, IsIn, Slice, BinOp, UnaryOp,
     Join, mean, var, std, Reduction, count, FloorDiv, UnaryStringFunction,
-    strlen, DateTime, Coerce, nunique, Distinct, By, Sort, Head, Label, Concat,
-    ReLabel, Merge, common_subexpression, Summary, Like, nelements, notnull
+    strlen, DateTime, Coerce, nunique, Distinct, By, Sort, Head, Tail, Label,
+    Concat, ReLabel, Merge, common_subexpression, Summary, Like, nelements,
+    notnull, Shift, BinaryMath, Pow,
 )
 
 from ..expr.broadcast import broadcast_collect
@@ -79,22 +80,78 @@ def inner_columns(s):
     raise TypeError()
 
 
+@dispatch(Projection, Select)
+def compute_up(expr, data, **kwargs):
+    d = dict((c.name, c) for c in getattr(data, 'inner_columns', data.c))
+    return data.with_only_columns([d[field] for field in expr.fields])
+
+
 @dispatch(Projection, Selectable)
-def compute_up(t, s, scope=None, **kwargs):
-    d = dict((c.name, c) for c in inner_columns(s))
-    return select(s).with_only_columns([d[field] for field in t.fields])
+def compute_up(expr, data, **kwargs):
+    return compute(expr, sa.select([data]), post_compute=False)
 
 
-@dispatch((Field, Projection), Select)
+@dispatch(Projection, sa.Column)
+def compute_up(expr, data, **kwargs):
+    selectables = [
+        compute(expr._child[field], data, post_compute=False)
+        for field in expr.fields
+    ]
+    froms = set(concat(s.froms for s in selectables))
+    assert 1 <= len(froms) <= 2
+    result = unify_froms(sa.select(first(sel.inner_columns)
+                                   for sel in selectables),
+                         froms)
+    return result.where(unify_wheres(selectables))
+
+
+@dispatch(Projection, Select)
+def compute_up(expr, data, **kwargs):
+    return data.with_only_columns(
+        first(compute(expr._child[field], data,
+                      post_compute=False).inner_columns)
+        for field in expr.fields
+    )
+
+
+@dispatch(Field, FromClause)
 def compute_up(t, s, **kwargs):
-    cols = list(s.inner_columns)
-    cols = [lower_column(cols[t._child.fields.index(c)]) for c in t.fields]
-    return s.with_only_columns(cols)
+    return s.c[t._name]
 
 
-@dispatch(Field, ClauseElement)
+def unify_froms(select, selectables):
+    return reduce(lambda x, y: x.select_from(y), selectables, select)
+
+
+def unify_wheres(selectables):
+    clauses = list(unique((s._whereclause for s in selectables
+                           if hasattr(s, '_whereclause')), key=str))
+    return reduce(and_, clauses) if clauses else None
+
+
+@dispatch(Field, Select)
+def compute_up(expr, data, **kwargs):
+    name = expr._name
+    try:
+        inner_columns = list(data.inner_columns)
+        names = list(c.name for c in data.inner_columns)
+        column = inner_columns[names.index(name)]
+    except (KeyError, ValueError):
+        single_column_select = compute(expr, first(data.inner_columns),
+                                       post_compute=False)
+        column = first(single_column_select.inner_columns)
+        result = unify_froms(sa.select([column]),
+                             data.froms + single_column_select.froms)
+        return result.where(unify_wheres([data, single_column_select]))
+    else:
+        return data.with_only_columns([column])
+
+
+@dispatch(Field, sa.Column)
 def compute_up(t, s, **kwargs):
-    return s.c.get(t._name)
+    assert len(s.foreign_keys) == 1, 'exactly one foreign key allowed'
+    key_col = first(s.foreign_keys).column
+    return sa.select([key_col.table.c[t._name]]).where(s == key_col)
 
 
 @dispatch(Broadcast, Select)
@@ -153,12 +210,101 @@ def compute_up(t, data, **kwargs):
         return t.op(t.lhs, column)
 
 
-@compute_up.register(BinOp, (ColumnElement, base), ColumnElement)
-@compute_up.register(BinOp, ColumnElement, (ColumnElement, base))
+@compute_up.register(
+    BinOp, (Select, ColumnElement, base), (Select, ColumnElement),
+)
+@compute_up.register(BinOp, (Select, ColumnElement), base)
 def binop_sql(t, lhs, rhs, **kwargs):
+    if isinstance(lhs, Select):
+        assert len(lhs.c) == 1, (
+            'Select cannot have more than a single column when doing'
+            ' arithmetic, got %s' % lhs
+        )
+        lhs = first(lhs.inner_columns)
+    if isinstance(rhs, Select):
+        assert len(rhs.c) == 1, (
+            'Select cannot have more than a single column when doing'
+            ' arithmetic, got %s' % rhs
+        )
+        rhs = first(rhs.inner_columns)
+
     return t.op(lhs, rhs)
 
 
+@dispatch(Pow, ColumnElement)
+def compute_up(t, data, **kwargs):
+    if isinstance(t.lhs, Expr):
+        return sa.func.pow(data, t.rhs)
+    else:
+        return sa.func.pow(t.lhs, data)
+
+
+@dispatch(Pow, Select)
+def compute_up(t, data, **kwargs):
+    assert len(data.c) == 1, \
+        'Select cannot have more than a single column when doing arithmetic'
+    column = first(data.inner_columns)
+    if isinstance(t.lhs, Expr):
+        return sa.func.pow(column, t.rhs)
+    else:
+        return sa.func.pow(t.lhs, column)
+
+
+@compute_up.register(Pow, (ColumnElement, base), ColumnElement)
+@compute_up.register(Pow, ColumnElement, base)
+def binop_sql_pow(t, lhs, rhs, **kwargs):
+    return sa.func.pow(lhs, rhs)
+
+
+@dispatch(BinaryMath, ColumnElement)
+def compute_up(t, data, **kwargs):
+    op = getattr(sa.func, type(t).__name__)
+    if isinstance(t.lhs, Expr):
+        return op(data, t.rhs)
+    else:
+        return op(t.lhs, data)
+
+
+@dispatch(BinaryMath, Select)
+def compute_up(t, data, **kwargs):
+    assert len(data.c) == 1, \
+        'Select cannot have more than a single column when doing arithmetic'
+    column = first(data.inner_columns)
+    op = getattr(sa.func, type(t).__name__)
+    if isinstance(t.lhs, Expr):
+        return op(column, t.rhs)
+    else:
+        return op(t.lhs, column)
+
+
+@compute_up.register(BinaryMath, (ColumnElement, base), ColumnElement)
+@compute_up.register(BinaryMath, ColumnElement, base)
+def binary_math_sql(t, lhs, rhs, **kwargs):
+    return getattr(sa.func, type(t).__name__)(lhs, rhs)
+
+
+@compute_up.register(BinaryMath, Select, base)
+def binary_math_sql_select(t, lhs, rhs, **kwargs):
+    left, right = first(lhs.inner_columns), rhs
+    result = getattr(sa.func, type(t).__name__)(left, right)
+    return reconstruct_select([result], lhs)
+
+
+@compute_up.register(BinaryMath, base, Select)
+def binary_math_sql_select(t, lhs, rhs, **kwargs):
+    left, right = lhs, first(rhs.inner_columns)
+    result = getattr(sa.func, type(t).__name__)(left, right)
+    return reconstruct_select([result], rhs)
+
+
+@compute_up.register(BinaryMath, Select, Select)
+def binary_math_sql_select(t, lhs, rhs, **kwargs):
+    left, right = first(lhs.inner_columns), first(rhs.inner_columns)
+    result = getattr(sa.func, type(t).__name__)(left, right)
+    assert lhs.table == rhs.table
+    return reconstruct_select([result], lhs.table)
+
+
 @dispatch(FloorDiv, ColumnElement)
 def compute_up(t, data, **kwargs):
     if isinstance(t.lhs, Expr):
@@ -168,7 +314,7 @@ def compute_up(t, data, **kwargs):
 
 
 @compute_up.register(FloorDiv, (ColumnElement, base), ColumnElement)
-@compute_up.register(FloorDiv, ColumnElement, (ColumnElement, base))
+@compute_up.register(FloorDiv, ColumnElement, base)
 def binop_sql(t, lhs, rhs, **kwargs):
     return sa.func.floor(lhs / rhs)
 
@@ -187,19 +333,47 @@ def compute_up(t, s, **kwargs):
 @dispatch(Selection, sa.sql.ColumnElement)
 def compute_up(expr, data, scope=None, **kwargs):
     predicate = compute(expr.predicate, data, post_compute=False)
-    return sa.select([data]).where(predicate)
+    return compute(
+        expr,
+        {expr._child: data, expr.predicate: predicate},
+        **kwargs
+    )
+
+
+@dispatch(Selection, sa.sql.ColumnElement, ColumnElement)
+def compute_up(expr, col, predicate, **kwargs):
+    return sa.select([col]).where(predicate)
 
 
 @dispatch(Selection, Selectable)
-def compute_up(t, s, scope=None, **kwargs):
-    ns = dict((t._child[col.name], col)
-              for col in getattr(s, 'inner_columns', s.columns))
-    predicate = compute(t.predicate, toolz.merge(ns, scope),
-                        optimize=False, post_compute=False)
+def compute_up(expr, sel, scope=None, **kwargs):
+    return compute(
+        expr,
+        {
+            expr._child: sel,
+            expr.predicate: compute(
+                expr.predicate,
+                toolz.merge(
+                    {
+                        expr._child[col.name]: col
+                        for col in getattr(sel, 'inner_columns', sel.columns)
+                    },
+                    scope,
+                ),
+                optimize=False,
+                post_compute=False,
+            ),
+        },
+        **kwargs
+    )
+
+
+@dispatch(Selection, Selectable, ColumnElement)
+def compute_up(expr, tbl, predicate, scope=None, **kwargs):
     try:
-        return s.where(predicate)
+        return tbl.where(predicate)
     except AttributeError:
-        return select([s]).where(predicate)
+        return select([tbl]).where(predicate)
 
 
 def select(s):
@@ -251,7 +425,6 @@ def _join_selectables(a, b, condition=None, **kwargs):
                                 a.join(b.froms[0], condition, **kwargs))
 
 
-
 @dispatch(ClauseElement, ClauseElement)
 def _join_selectables(a, b, condition=None, **kwargs):
     return a.join(b, condition, **kwargs)
@@ -382,7 +555,7 @@ def reconstruct_select(columns, original, **kwargs):
 def compute_up(expr, data, **kwargs):
     if expr.axis != (0,):
         raise ValueError('axis not equal to 0 not defined for SQL reductions')
-    data = data.cte(name=next(aliases))
+    data = data.alias(name=next(aliases))
     cols = list(inner_columns(data))
     d = dict((expr._child[c], cols[i])
              for i, c in enumerate(expr._child.fields))
@@ -469,7 +642,7 @@ def compute_up(t, s, **kwargs):
     return select([list(inner_columns(result))[0].label(t._name)])
 
 
-@dispatch(nunique, sa.Column)
+@dispatch(nunique, (sa.sql.elements.Label, sa.Column))
 def compute_up(t, s, **kwargs):
     if t.axis != (0,):
         raise ValueError('axis not equal to 0 not defined for SQL reductions')
@@ -482,26 +655,35 @@ def compute_up(expr, data, **kwargs):
 
 
 @dispatch(By, sa.Column)
-def compute_up(expr, data, **kwargs):
-    grouper = lower_column(data)
-    app = expr.apply
-    if isinstance(app, Reduction):
-        reductions = [compute(app, data, post_compute=False)]
-    elif isinstance(app, Summary):
-        reductions = [compute(val, data, post_compute=None).label(name)
-                      for val, name in zip(app.values, app.fields)]
+def compute_up(expr, data, scope=None, **kwargs):
+    data = lower_column(data)
+    grouper = compute(expr.grouper, scope, post_compute=False, **kwargs)
 
-    return sa.select([grouper] + reductions).group_by(grouper)
+    app = expr.apply
+    reductions = [compute(val, data, post_compute=None).label(name)
+                  for val, name in zip(app.values, app.fields)]
+
+    froms = list(unique(chain(get_all_froms(grouper),
+                              concat(map(get_all_froms, reductions)))))
+    inner_cols = list(getattr(grouper, 'inner_columns', [grouper]))
+    grouper_cols = inner_cols[:]
+    inner_cols.extend(concat(
+        getattr(getattr(r, 'element', None), 'inner_columns', [r])
+        for r in reductions
+    ))
+    wheres = unify_wheres([grouper] + reductions)
+    sel = unify_froms(sa.select(inner_cols, whereclause=wheres), froms)
+    return sel.group_by(*grouper_cols)
 
 
 @dispatch(By, ClauseElement)
 def compute_up(expr, data, **kwargs):
-    if not valid_grouper(expr):
+    if not valid_grouper(expr.grouper):
         raise TypeError("Grouper must have a non-nested record or one "
                         "dimensional collection datashape, "
                         "got %s of type %r with dshape %s" %
                         (expr.grouper, type(expr.grouper).__name__,
-                         expr.dshape))
+                         expr.grouper.dshape))
     grouper = get_inner_columns(compute(expr.grouper, data,
                                         post_compute=False))
     app = expr.apply
@@ -573,14 +755,14 @@ def is_nested_record(measure):
     if not isrecord(measure):
         raise TypeError('Input must be a Record type got %s of type %r' %
                         (measure, type(measure).__name__))
-    return not all(isscalar(t) for t in measure.types)
+    return not all(isscalar(getattr(t, 'key', t)) for t in measure.types)
 
 
 def valid_grouper(expr):
     ds = expr.dshape
     measure = ds.measure
     return (iscollection(ds) and
-            (isscalar(measure) or
+            (isscalar(getattr(measure, 'key', measure)) or
              (isrecord(measure) and not is_nested_record(measure))))
 
 
@@ -594,11 +776,12 @@ def valid_reducer(expr):
 
 @dispatch(By, Select)
 def compute_up(expr, data, **kwargs):
-    if not valid_grouper(expr):
+    if not valid_grouper(expr.grouper):
         raise TypeError("Grouper must have a non-nested record or one "
                         "dimensional collection datashape, "
                         "got %s of type %r with dshape %s" %
-                        (expr.grouper, type(expr.grouper).__name__, expr.dshape))
+                        (expr.grouper, type(expr.grouper).__name__,
+                         expr.grouper.dshape))
 
     s = alias_it(data)
 
@@ -674,7 +857,7 @@ def compute_up(t, s, **kwargs):
     assert len(s.c) == 1, \
         'expected %s to have a single column but has %d' % (s, len(s.c))
     inner_column, = s.inner_columns
-    return reconstruct_select([inner_column.label(t.label)], s).as_scalar()
+    return reconstruct_select([inner_column.label(t.label)], s)
 
 
 @dispatch(Expr, ScalarSelect)
@@ -745,6 +928,21 @@ def get_all_froms(t):
     return [t]
 
 
+@dispatch(sa.sql.elements.ColumnElement)
+def get_all_froms(colelement):
+    return list(unique(concat(map(get_all_froms, colelement.get_children()))))
+
+
+@dispatch((ScalarSelect, sa.sql.elements.Label))
+def get_all_froms(element):
+    return get_all_froms(element.element)
+
+
+@dispatch(sa.sql.functions.FunctionElement)
+def get_all_froms(function):
+    return list(unique(concat(map(get_all_froms, function.clauses.clauses))))
+
+
 @dispatch(ColumnClause)
 def get_all_froms(c):
     return [c.table]
@@ -801,17 +999,9 @@ def compute_up(t, s, **kwargs):
     )
 
 
-@dispatch(Like, Selectable)
+@dispatch(Like, ColumnElement)
 def compute_up(t, s, **kwargs):
-    return compute_up(t, select(s), **kwargs)
-
-
-@dispatch(Like, Select)
-def compute_up(t, s, **kwargs):
-    items = [(f.c.get(name), pattern.replace('*', '%'))
-             for name, pattern in t.patterns.items()
-             for f in s.froms if name in f.c]
-    return s.where(reduce(and_, [key.like(pattern) for key, pattern in items]))
+    return s.like(t.pattern.replace('*', '%').replace('?', '_'))
 
 
 string_func_names = {
@@ -866,7 +1056,7 @@ def compute_up(expr, data, **kwargs):
     return table_of_engine(data, expr._name)
 
 
-@dispatch(DateTime, (ClauseElement, sa.sql.elements.ColumnElement))
+@dispatch(DateTime, ColumnElement)
 def compute_up(expr, data, **kwargs):
     if expr.attr == 'date':
         return sa.func.date(data).label(expr._name)
@@ -897,9 +1087,45 @@ def engine_of(x):
     raise NotImplementedError("Can't deterimine engine of %s" % x)
 
 
+@dispatch(Expr)
+def _subexpr_optimize(expr):
+    return expr
+
+
+@dispatch(Tail)
+def _subexpr_optimize(expr):
+    child = sorter = expr._child
+    while not isinstance(sorter, Sort):
+        try:
+            sorter = sorter._child
+        except AttributeError:
+            break
+    else:
+        # Invert the sort order, then take the head, then re-sort based on
+        # the original key.
+        return child._subs({
+            sorter: sorter._child.sort(
+                sorter._key,
+                ascending=not sorter.ascending,
+            ),
+        }).head(expr.n).sort(sorter._key, ascending=sorter.ascending)
+
+    # If there is no sort order, then we can swap out a head with a tail.
+    # This is equivalent in this backend and considerably faster.
+    warnings.warn(
+        "'tail' of a sql operation with no sort is the same as 'head'",
+    )
+    return child.head(expr.n)
+
+
 @dispatch(Expr, ClauseElement)
 def optimize(expr, _):
-    return broadcast_collect(expr)
+    collected = broadcast_collect(expr, no_recurse=Selection)
+    return reduce(
+        lambda expr, term: expr._subs({term: _subexpr_optimize(term)}),
+        collected._subterms(),
+        collected,
+    )
 
 
 @dispatch(Field, sa.MetaData)
@@ -957,3 +1183,15 @@ def compute_up(expr, data, **kwargs):
 @dispatch(Coerce, ColumnElement)
 def compute_up(expr, data, **kwargs):
     return sa.cast(data, dshape_to_alchemy(expr.to)).label(expr._name)
+
+
+@dispatch(Coerce, Select)
+def compute_up(expr, data, **kwargs):
+    column = first(data.inner_columns)
+    cast = sa.cast(column, dshape_to_alchemy(expr.to)).label(expr._name)
+    return reconstruct_select([cast], data)
+
+
+@dispatch(Shift, ColumnElement)
+def compute_up(expr, data, **kwargs):
+    return sa.func.lag(data, expr.n).over().label(expr._name)
