@@ -5,22 +5,26 @@ import pytest
 sa = pytest.importorskip('sqlalchemy')
 
 import itertools
-import re
+import sqlite3
 from distutils.version import LooseVersion
 
 
 import datashape
 from odo import into, resource, discover
+import numpy as np
+import pandas as pd
+import pandas.util.testing as tm
 from pandas import DataFrame
 from toolz import unique
 
+from odo import odo, drop
 from blaze.compute.sql import compute, select, lower_column, compute_up
 from blaze.expr import (
     symbol, transform, summary, by, sin, join,
     floor, cos, merge, nunique, mean, sum, count, exp
 )
 from blaze.compatibility import xfail
-from blaze.utils import tmpfile, example
+from blaze.utils import tmpfile, example, normalize
 
 
 def computefull(t, s):
@@ -97,12 +101,6 @@ sbig = sa.Table('accountsbig', metadata,
                 sa.Column('id', sa.Integer, primary_key=True))
 
 
-def normalize(s):
-    s = ' '.join(s.strip().split()).lower()
-    s = re.sub(r'(alias)_?\d*', r'\1', s)
-    return re.sub(r'__([A-Za-z_][A-Za-z_0-9]*)', r'\1', s)
-
-
 def test_table():
     result = str(computefull(t, s))
     expected = """
@@ -156,7 +154,6 @@ def test_arithmetic():
         str(sa.select([s.c.amount + s.c.id * 2]))
 
 
-
 def test_join():
     metadata = sa.MetaData()
     lhs = sa.Table('amounts', metadata,
@@ -207,6 +204,7 @@ def test_join():
     order by
         anon_1.amount asc""")
 
+
 def test_clean_complex_join():
     metadata = sa.MetaData()
     lhs = sa.Table('amounts', metadata,
@@ -341,17 +339,17 @@ def test_count_on_table():
     result = compute(t[t.amount > 0].count(), s)
     assert (
         normalize(str(result)) == normalize("""
-        SELECT count(accounts.id) as count_1
+        SELECT count(accounts.id) as t_count
         FROM accounts
         WHERE accounts.amount > :amount_1""")
 
         or
 
         normalize(str(result)) == normalize("""
-        SELECT count(alias.id) as count
+        SELECT count(alias.id) as t_count
         FROM (SELECT accounts.name AS name, accounts.amount AS amount, accounts.id AS id
               FROM accounts
-              WHERE accounts.amount > :amount_1) as alias"""))
+              WHERE accounts.amount > :amount_1) as alias_2"""))
 
 
 def test_distinct():
@@ -407,17 +405,6 @@ def test_by_head():
     t2 = t.head(100)
     expr = by(t2['name'], total=t2['amount'].sum())
     result = compute(expr, s)
-    # s2 = select(s).limit(100)
-    # expected = sa.select([s2.c.name,
-    #                       sa.sql.functions.sum(s2.c.amount).label('amount_sum')]
-    #                      ).group_by(s2.c.name)
-    expected = """
-    SELECT alias.name, sum(alias.amount) as total
-    FROM (SELECT accounts.name AS name, accounts.amount AS amount, accounts.id AS ID
-          FROM accounts
-          LIMIT :param_1) as alias
-    GROUP BY alias.name"""
-
     expected = """
     SELECT accounts.name, sum(accounts.amount) as total
     FROM accounts
@@ -733,14 +720,23 @@ def test_clean_join():
     assert (normalize(str(result)) == normalize(expected1) or
             normalize(str(result)) == normalize(expected2))
 
+
 def test_like():
-    expr = t.like(name='Alice*')
+    expr = t[t.name.like('Alice*')]
     assert normalize(str(compute(expr, s))) == normalize("""
     SELECT accounts.name, accounts.amount, accounts.id
     FROM accounts
     WHERE accounts.name LIKE :name_1""")
 
 
+def test_not_like():
+    expr = t[~t.name.like('Alice*')]
+    assert normalize(str(compute(expr, s))) == normalize("""
+    SELECT accounts.name, accounts.amount, accounts.id
+    FROM accounts
+    WHERE accounts.name NOT LIKE :name_1""")
+
+
 def test_strlen():
     expr = t.name.strlen()
     result = str(compute(expr, s))
@@ -759,15 +755,19 @@ def test_columnwise_on_complex_selection():
 
 
 def test_reductions_on_complex_selections():
-    assert normalize(str(select(compute(t[t.amount > 0].id.sum(), s)))) == \
-        normalize("""
-    with alias as
-        (select accounts.id as id
-         from
-            accounts
-         where
-            accounts.amount > :amount_1)
-    select sum(alias.id) as id_sum from alias""")
+    assert (
+        normalize(str(select(compute(t[t.amount > 0].id.sum(), s)))) ==
+        normalize(
+            """
+            select
+                sum(alias.id) as id_sum
+            from (select
+                    accounts.id as id
+                  from accounts
+                  where accounts.amount > :amount_1) as alias
+            """
+        )
+    )
 
 
 def test_clean_summary_by_where():
@@ -948,16 +948,16 @@ def test_join_on_same_table():
     result = compute(expr, {t: T})
 
     assert normalize(str(result)) == normalize("""
-    with alias as
-    (select
-         tab_left.b as b_left
-     from
-         tab as tab_left
-     join
-         tab as tab_right
-     on
-         tab_left.a = tab_right.a)
-    select sum(alias.b_left) as b_left_sum from alias""")
+    select sum(alias.b_left) as b_left_sum from
+        (select
+            tab_left.b as b_left
+        from
+            tab as tab_left
+        join
+            tab as tab_right
+        on
+            tab_left.a = tab_right.a) as
+        alias""")
 
     expr = join(t, t, 'a')
     expr = summary(total=expr.a.sum(), smallest=expr.b_right.min())
@@ -1586,11 +1586,11 @@ def test_transform_then_project():
 def test_reduce_does_not_compose():
     expr = by(t.name, counts=t.count()).counts.max()
     result = str(compute(expr, s))
-    expected = """WITH alias AS
-(SELECT count(accounts.id) AS counts
-FROM accounts GROUP BY accounts.name)
- SELECT max(alias.counts) AS counts_max
-FROM alias"""
+    expected = """
+    SELECT max(alias.counts) AS counts_max
+    FROM
+    (SELECT count(accounts.id) AS counts
+    FROM accounts GROUP BY accounts.name) as alias"""
     assert normalize(result) == normalize(expected)
 
 
@@ -1854,3 +1854,103 @@ def test_distinct_on_label():
     FROM accounts
     """
     assert normalize(str(result)) == normalize(expected)
+
+
+@pytest.mark.parametrize('n', [-1, 0, 1])
+def test_shift_on_column(n):
+    expr = t.name.shift(n)
+    result = compute(expr, s)
+    expected = """SELECT
+        lag(accounts.name, :lag_1) over () as name
+    FROM accounts
+    """
+    assert normalize(str(result)) == normalize(expected)
+
+
+@pytest.fixture
+def lhs():
+    return pd.DataFrame.from_records(
+        dict(zip('cd', row)) for row in zip(
+            ['AAPL', 'FB', 'AMZN', 'FB', 'GOOG'] * 2,
+            list(np.random.rand(10) * 100)
+        )
+    )
+
+
+@pytest.yield_fixture
+def rhs():
+    try:
+        t = resource(
+            'sqlite:///:memory:::rhs',
+            dshape='var * {a: int64, b: ?float64, c: ?string}'
+        )
+    except sa.exc.OperationalError as e:
+        pytest.skip(str(e))
+    else:
+        try:
+            t.insert().values([
+                dict(zip(t.columns.keys(), row)) for row in zip(
+                    range(5),
+                    list(np.random.rand(5) * 100),
+                    ['AAPL', 'AAPL', 'MSFT', 'FB', 'MSFT']
+                )
+            ]).execute()
+            yield t
+        finally:
+            drop(t)
+
+
+
+@pytest.mark.skipif(LooseVersion(sqlite3.version) <= '2.6.0',
+                    reason="Older sqlite3 versions don't support this")
+def test_multiple_table_join(lhs, rhs):
+    lexpr = symbol('lhs', discover(lhs))
+    rexpr = symbol('rhs', discover(rhs))
+    expr = join(lexpr, rexpr, 'c')
+    result = compute(expr, {lexpr: lhs, rexpr: rhs})
+    expected = pd.merge(lhs, odo(rhs, pd.DataFrame), on='c')
+    tm.assert_frame_equal(result, expected)
+
+
+def test_empty_string_comparison_with_option_type():
+    expr = nt.amount[nt.name == '']
+    result = compute(expr, s)
+    expected = """
+    SELECT accounts.amount
+    FROM accounts
+    WHERE accounts.name = :name_1
+    """
+    assert normalize(str(result)) == normalize(expected)
+
+
+def test_tail_no_sort():
+    assert (
+        normalize(str(compute(t.head(), {t: s}))) ==
+        normalize(str(compute(t.tail(), {t: s})))
+    )
+
+
+def test_tail_of_sort():
+    expected = normalize(str(compute(
+        t.sort('id', ascending=False).head(5).sort('id'),
+        {t: s},
+    )))
+    result = normalize(str(compute(t.sort('id').tail(5), {t: s})))
+    assert expected == result
+
+
+def test_tail_sort_in_chilren():
+    expected = normalize(str(compute(
+        t.name.sort('id', ascending=False).head(5).sort('id'),
+        {t: s},
+    )))
+    result = normalize(str(compute(t.name.sort('id').tail(5), {t: s})))
+    assert expected == result
+
+
+def test_selection_inner_inputs():
+    result = normalize(str(compute(t[t.id == tdate.id], {t: s, tdate: sdate})))
+    expected = normalize("""
+    select {a}.name, {a}.amount, {a}.id from {a}, {b} where {a}.id = {b}.id
+    """).format(a=s.name, b=sdate.name)
+    assert result == expected
