@@ -12,20 +12,44 @@ import toolz
 from toolz import concat, memoize, partial, first
 from toolz.curried import map, filter
 
+from datashape import dshape, DataShape, Record, Var, Mono, Fixed
+from datashape.predicates import isscalar, iscollection, isboolean, isrecord
+
+from ..compatibility import _strtypes, builtins, boundmethod
 from ..compatibility import _strtypes, builtins, boundmethod, PY2
 from .core import Node, subs, common_subexpression, path
 from .method_dispatch import select_functions
 from ..dispatch import dispatch
 from .utils import hashable_index, replace_slices
+from ..utils import weakmemoize
 
 
-__all__ = ['Expr', 'ElemWise', 'Field', 'Symbol', 'discover', 'Projection',
-           'projection', 'Selection', 'selection', 'Label', 'label', 'Map',
-           'ReLabel', 'relabel', 'Apply', 'apply', 'Slice', 'shape', 'ndim',
-           'label', 'symbol', 'Coerce', 'coerce']
-
-
-_attr_cache = dict()
+__all__ = [
+    'Apply',
+    'Coerce',
+    'ElemWise',
+    'Expr',
+    'Field',
+    'Label',
+    'Map',
+    'Projection',
+    'ReLabel',
+    'Selection',
+    'SimpleSelection',
+    'Slice',
+    'Symbol',
+    'apply',
+    'coerce',
+    'discover',
+    'label',
+    'label',
+    'ndim',
+    'projection',
+    'relabel',
+    'selection',
+    'shape',
+    'symbol',
+]
 
 
 def isvalid_identifier(s):
@@ -67,7 +91,7 @@ def valid_identifier(s):
     >>> valid_identifier('1a')
     """
     if isinstance(s, _strtypes):
-        if s[0].isdigit():
+        if not s or s[0].isdigit():
             return
         return s.replace(' ', '_').replace('.', '_').replace('-', '_')
     return s
@@ -81,8 +105,10 @@ class Expr(Node):
     contains shared logic and syntax.  It in turn inherits from ``Node`` which
     holds all tree traversal logic
     """
+    __slots__ = '_hash', '__weakref__', '__dict__'
+
     def _get_field(self, fieldname):
-        if not isinstance(self.dshape.measure, Record):
+        if not isinstance(self.dshape.measure, (Record, datashape.Map)):
             if fieldname == self._name:
                 return self
             raise ValueError(
@@ -95,8 +121,8 @@ class Expr(Node):
             return self._get_field(key)
         elif isinstance(key, Expr) and iscollection(key.dshape):
             return selection(self, key)
-        elif (isinstance(key, list)
-                and builtins.all(isinstance(k, _strtypes) for k in key)):
+        elif (isinstance(key, list) and
+              builtins.all(isinstance(k, _strtypes) for k in key)):
             if set(key).issubset(self.fields):
                 return self._project(key)
             else:
@@ -117,13 +143,31 @@ class Expr(Node):
         return projection(self, key)
 
     @property
+    @weakmemoize
     def schema(self):
+        try:
+            m = self._schema
+        except AttributeError:
+            pass
+        else:
+            return m()
+
         return datashape.dshape(self.dshape.measure)
 
+    @property
+    @weakmemoize
+    def dshape(self):
+        return self._dshape()
+
     @property
     def fields(self):
         if isinstance(self.dshape.measure, Record):
             return self.dshape.measure.names
+        elif isinstance(self.dshape.measure, datashape.Map):
+            if not isrecord(self.dshape.measure.value):
+                raise TypeError('Foreign key must reference a '
+                                'Record datashape')
+            return self.dshape.measure.value.names
         name = getattr(self, '_name', None)
         if name is not None:
             return [self._name]
@@ -148,27 +192,23 @@ class Expr(Node):
 
     def __dir__(self):
         result = dir(type(self))
-        if isrecord(self.dshape.measure) and self.fields:
-            result.extend(list(map(valid_identifier, self.fields)))
+        if (isrecord(self.dshape.measure) or
+            isinstance(self.dshape.measure, datashape.Map) and
+                self.fields):
+            result.extend(map(valid_identifier, self.fields))
 
-        d = toolz.merge(schema_methods(self.dshape.measure),
-                        dshape_methods(self.dshape))
-        result.extend(list(d))
+        result.extend(toolz.merge(schema_methods(self.dshape.measure),
+                                  dshape_methods(self.dshape)))
 
         return sorted(set(filter(isvalid_identifier, result)))
 
     def __getattr__(self, key):
-        if key == '_hash':
-            raise AttributeError()
-        try:
-            return _attr_cache[(self, key)]
-        except:
-            pass
+        assert key != '_hash', \
+            '%s expressions should set _hash in __init__' % type(self).__name__
         try:
             result = object.__getattribute__(self, key)
         except AttributeError:
-            fields = dict(zip(map(valid_identifier, self.fields),
-                              self.fields))
+            fields = dict(zip(map(valid_identifier, self.fields), self.fields))
 
             # prefer the method if there's a field with the same name
             methods = toolz.merge(
@@ -188,15 +228,19 @@ class Expr(Node):
                     result = self[fields[key]]
             else:
                 raise
-        _attr_cache[(self, key)] = result
+
+        # cache the attribute lookup, getattr will not be invoked again.
+        setattr(self, key, result)
         return result
 
     @property
     def _name(self):
-        if (isscalar(self.dshape.measure) and
-                len(self._inputs) == 1 and
-                isscalar(self._child.dshape.measure)):
-            return self._child._name
+        measure = self.dshape.measure
+        if len(self._inputs) == 1 and isscalar(getattr(measure, 'key',
+                                                       measure)):
+            child_measure = self._child.dshape.measure
+            if isscalar(getattr(child_measure, 'key', child_measure)):
+                return self._child._name
 
     def __enter__(self):
         """ Enter context """
@@ -233,11 +277,6 @@ def _symbol_key(args, kwargs):
     return (name, ds, token)
 
 
-@memoize(cache=_symbol_cache, key=_symbol_key)
-def symbol(name, dshape, token=None):
-    return Symbol(name, dshape, token=token)
-
-
 class Symbol(Expr):
     """
     Symbolic data.  The leaf of a Blaze expression
@@ -261,6 +300,7 @@ class Symbol(Expr):
             dshape = DataShape(dshape)
         self.dshape = dshape
         self._token = token
+        self._hash = None
 
     def __str__(self):
         return self._name or ''
@@ -269,6 +309,12 @@ class Symbol(Expr):
         return dict()
 
 
+@memoize(cache=_symbol_cache, key=_symbol_key)
+@copydoc(Symbol)
+def symbol(name, dshape, token=None):
+    return Symbol(name, dshape, token=token)
+
+
 @dispatch(Symbol, dict)
 def _subs(o, d):
     """ Subs symbols using symbol function
@@ -284,10 +330,10 @@ class ElemWise(Expr):
 
     The shape of this expression matches the shape of the child.
     """
-    @property
-    def dshape(self):
-        return datashape.DataShape(*(self._child.dshape.shape
-                                     + tuple(self.schema)))
+    def _dshape(self):
+        return datashape.DataShape(
+            *(self._child.dshape.shape + tuple(self.schema))
+        )
 
 
 class Field(ElemWise):
@@ -319,10 +365,12 @@ class Field(ElemWise):
     def _expr(self):
         return symbol(self._name, datashape.DataShape(self.dshape.measure))
 
-    @property
-    def dshape(self):
+    def _dshape(self):
         shape = self._child.dshape.shape
-        schema = self._child.dshape.measure.dict[self._name]
+        measure = self._child.dshape.measure
+
+        # TODO: is this too special-case-y?
+        schema = getattr(measure, 'value', measure).dict[self._name]
 
         shape = shape + schema.shape
         schema = (schema.measure,)
@@ -351,10 +399,10 @@ class Projection(ElemWise):
     def fields(self):
         return list(self._fields)
 
-    @property
-    def schema(self):
-        d = self._child.schema[0].dict
-        return DataShape(Record([(name, d[name]) for name in self.fields]))
+    def _schema(self):
+        measure = self._child.schema.measure
+        d = getattr(measure, 'value', measure).dict
+        return DataShape(Record((name, d[name]) for name in self.fields))
 
     def __str__(self):
         return '%s[%s]' % (self._child, self.fields)
@@ -428,8 +476,7 @@ class Slice(Expr):
     """
     __slots__ = '_hash', '_child', '_index'
 
-    @property
-    def dshape(self):
+    def _dshape(self):
         return self._child.dshape.subshape[self.index]
 
     @property
@@ -455,17 +502,28 @@ class Selection(Expr):
     >>> deadbeats = accounts[accounts.amount < 0]
     """
     __slots__ = '_hash', '_child', 'predicate'
+    __inputs__ = '_child', 'predicate'
+
+    @property
+    def _name(self):
+        return self._child._name
 
     def __str__(self):
         return "%s[%s]" % (self._child, self.predicate)
 
-    @property
-    def dshape(self):
+    def _dshape(self):
         shape = list(self._child.dshape.shape)
         shape[0] = Var()
         return DataShape(*(shape + [self._child.dshape.measure]))
 
 
+class SimpleSelection(Selection):
+    """Internal selection class that does not treat the predicate as an input.
+    """
+    __slots__ = Selection.__slots__
+    __inputs__ = '_child',
+
+
 @copydoc(Selection)
 def selection(table, predicate):
     subexpr = common_subexpression(table, predicate)
@@ -505,8 +563,7 @@ class Label(ElemWise):
     """
     __slots__ = '_hash', '_child', 'label'
 
-    @property
-    def schema(self):
+    def _schema(self):
         return self._child.schema
 
     @property
@@ -572,8 +629,7 @@ class ReLabel(ElemWise):
     """
     __slots__ = '_hash', '_child', 'labels'
 
-    @property
-    def schema(self):
+    def _schema(self):
         subs = dict(self.labels)
         param = self._child.dshape.measure.parameters[0]
         return DataShape(Record([[subs.get(name, name), dtype]
@@ -630,12 +686,11 @@ class Map(ElemWise):
     --------
     blaze.expr.expresions.Apply
     """
-    __slots__ = '_hash', '_child', 'func', '_schema', '_name0'
+    __slots__ = '_hash', '_child', 'func', '_asschema', '_name0'
 
-    @property
-    def schema(self):
-        if self._schema:
-            return dshape(self._schema)
+    def _schema(self):
+        if self._asschema:
+            return dshape(self._asschema)
         else:
             raise NotImplementedError("Schema of mapped column not known.\n"
                                       "Please specify datashape keyword in "
@@ -694,18 +749,16 @@ class Apply(Expr):
 
     blaze.expr.expressions.Map
     """
-    __slots__ = '_hash', '_child', 'func', '_dshape', '_splittable'
+    __slots__ = '_hash', '_child', 'func', '_asdshape', '_splittable'
 
-    @property
-    def schema(self):
+    def _schema(self):
         if iscollection(self.dshape):
             return self.dshape.subshape[0]
         else:
             raise TypeError("Non-tabular datashape, %s" % self.dshape)
 
-    @property
-    def dshape(self):
-        return dshape(self._dshape)
+    def _dshape(self):
+        return dshape(self._asdshape)
 
 
 @copydoc(Apply)
@@ -713,7 +766,7 @@ def apply(expr, func, dshape, splittable=False):
     return Apply(expr, func, datashape.dshape(dshape), splittable)
 
 
-class Coerce(Expr):
+class Coerce(ElemWise):
     """Coerce an expression to a different type.
 
     Examples
@@ -728,14 +781,9 @@ class Coerce(Expr):
     """
     __slots__ = '_hash', '_child', 'to'
 
-    @property
-    def schema(self):
+    def _schema(self):
         return self.to
 
-    @property
-    def dshape(self):
-        return DataShape(*(self._child.shape + (self.schema,)))
-
     def __str__(self):
         return '%s.coerce(to=%r)' % (self._child, str(self.schema))
 
