@@ -2,6 +2,8 @@
 Dispatches functions like ``sum`` to builtins, numpy, or blaze depending on
 input
 
+>>> from blaze import sum, symbol
+
 >>> sum([1, 2, 3])
 6
 
@@ -21,15 +23,14 @@ from __future__ import absolute_import, division, print_function
 import math as pymath
 from numbers import Number
 import numpy as np
-from datashape.predicates import iscollection, isscalar
-from toolz import curry
 
 from multipledispatch import Dispatcher
-from ..dispatch import dispatch, namespace
+from ..dispatch import namespace
 from ..compatibility import builtins
 from . import reductions
 from . import math as blazemath
-from .expressions import Expr, symbol
+from .core import base
+from .expressions import Expr
 
 
 """
@@ -66,9 +67,11 @@ def sqrt(o):
 math_names = '''abs sqrt sin cos tan sinh cosh tanh acos acosh asin asinh atan atanh
 exp log expm1 log10 log1p radians degrees ceil floor trunc isnan'''.split()
 
+binary_math_names = "atan2 copysign fmod hypot ldexp greatest least".split()
+
 reduction_names = '''any all sum min max mean var std'''.split()
 
-__all__ = math_names + reduction_names
+__all__ = math_names + binary_math_names + reduction_names
 
 
 types = {builtins: object,
@@ -77,6 +80,27 @@ types = {builtins: object,
          blazemath: Expr,
          reductions: Expr}
 
+binary_types = {
+    builtins: [(object, object)],
+    np: [((np.ndarray, np.number), (np.ndarray, np.number))],
+    pymath: [(Number, Number)],
+    blazemath: [(Expr, (Expr, base)), (base, Expr)]
+}
+
+
+fallback_binary_mappings = {
+    'greatest': {
+        builtins: max,
+        np: np.maximum,
+        pymath: max,
+    },
+    'least': {
+        builtins: min,
+        np: np.minimum,
+        pymath: min,
+    }
+}
+
 
 for funcname in math_names:  # sin, sqrt, ceil, ...
     d = Dispatcher(funcname)
@@ -89,6 +113,21 @@ for funcname in math_names:  # sin, sqrt, ceil, ...
     locals()[funcname] = d
 
 
+for funcname in binary_math_names:  # hypot, atan2, fmod, ...
+    d = Dispatcher(funcname)
+
+    for module, pairs in binary_types.items():
+        for pair in pairs:
+            if hasattr(module, funcname):
+                d.add(pair, getattr(module, funcname))
+            elif funcname in fallback_binary_mappings:
+                assert module in fallback_binary_mappings[funcname], module.__name__
+                d.add(pair, fallback_binary_mappings[funcname][module])
+
+    namespace[funcname] = d
+    locals()[funcname] = d
+
+
 for funcname in reduction_names:  # any, all, sum, max, ...
     d = Dispatcher(funcname)
 
