@@ -4,16 +4,23 @@ import pytest
 pytest.importorskip('flask')
 
 from base64 import b64encode
+from contextlib import contextmanager
+from copy import copy
 
 import datashape
+from datashape.util.testing import assert_dshape_equal
 import numpy as np
+from odo import odo, convert
 from datetime import datetime
 from pandas import DataFrame
+from pandas.util.testing import assert_frame_equal
 from toolz import pipe
 
-from odo import odo
+from blaze.dispatch import dispatch
+from blaze.expr import Expr
 from blaze.utils import example
-from blaze import discover, symbol, by, CSV, compute, join, into, resource
+from blaze import (discover, symbol, by, CSV, compute, join, into, resource,
+                   Data)
 from blaze.server.client import mimetype
 from blaze.server.server import Server, to_tree, from_tree
 from blaze.server.serialization import all_formats
@@ -31,10 +38,43 @@ events = DataFrame([[1, datetime(2000, 1, 1, 12, 0, 0)],
 
 db = resource('sqlite:///' + example('iris.db'))
 
-data = {'accounts': accounts,
-          'cities': cities,
-          'events': events,
-              'db': db}
+
+class DumbResource(object):
+    df = DataFrame({
+        'a': np.arange(5),
+        'b': np.arange(5, 10),
+    })
+
+    class NoResource(Exception):
+        pass
+
+
+@convert.register(DataFrame, DumbResource)
+def dumb_to_df(d, return_df=None, **kwargs):
+    if return_df is None:
+        raise DumbResource.NoResource('return_df must be passed')
+    to_return = odo(return_df, DataFrame, dshape=discover(d))
+    assert_frame_equal(to_return, DumbResource.df)
+    return to_return
+
+
+@dispatch(Expr, DumbResource)
+def compute_down(expr, d, **kwargs):
+    return dumb_to_df(d, **kwargs)
+
+
+@discover.register(DumbResource)
+def _discover_dumb(d):
+    return discover(DumbResource.df)
+
+
+data = {
+    'accounts': accounts,
+    'cities': cities,
+    'events': events,
+    'db': db,
+    'dumb': DumbResource(),
+}
 
 
 @pytest.fixture(scope='module')
@@ -43,6 +83,12 @@ def server():
     s.app.testing = True
     return s
 
+@contextmanager
+def temp_server(data=None):
+    """For when we want to mutate the server"""
+    s = Server(copy(data), formats=all_formats)
+    s.app.testing = True
+    yield s.app.test_client()
 
 @pytest.yield_fixture
 def test(server):
@@ -50,9 +96,29 @@ def test(server):
         yield c
 
 
+@pytest.yield_fixture
+def empty_server():
+    s = Server(formats=all_formats)
+    s.app.testing = True
+    with s.app.test_client() as c:
+        yield c
+
+
+@pytest.yield_fixture
+def iris_server():
+    iris = CSV(example('iris.csv'))
+    s = Server(iris, all_formats)
+    s.app.testing = True
+    with s.app.test_client() as c:
+        yield c
+
+
 def test_datasets(test):
     response = test.get('/datashape')
-    assert response.data.decode('utf-8') == str(discover(data))
+    assert_dshape_equal(
+            datashape.dshape(response.data.decode('utf-8')),
+            datashape.dshape(discover(data))
+            )
 
 
 @pytest.mark.parametrize('serial', all_formats)
@@ -183,15 +249,6 @@ def dont_test_compute_with_namespace(test, serial):
     assert data['names'] == ['name']
 
 
-@pytest.yield_fixture
-def iris_server():
-    iris = CSV(example('iris.csv'))
-    s = Server(iris, all_formats)
-    s.app.testing = True
-    with s.app.test_client() as c:
-        yield c
-
-
 iris = CSV(example('iris.csv'))
 
 
@@ -486,3 +543,193 @@ def test_minute_query(test, serial):
     }
     assert result.status_code == 200
     assert expected == serial.loads(result.data)
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_isin(test, serial):
+    expr = t.events.value.isin(frozenset([1]))
+    query = {'expr': to_tree(expr)}
+    result = test.post(
+        '/compute',
+        headers=mimetype(serial),
+        data=serial.dumps(query)
+    )
+    expected = {
+        'data': [True, False],
+        'names': ['value'],
+        'datashape': '2 * bool',
+    }
+    assert result.status_code == 200
+    assert expected == serial.loads(result.data)
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_add_data_to_empty_server(empty_server, serial):
+    # add data
+    with temp_server() as test:
+        iris_path = example('iris.csv')
+        blob = serial.dumps({'iris': iris_path})
+        response1 = empty_server.post(
+            '/add',
+            headers=mimetype(serial),
+            data=blob,
+        )
+        assert 'OK' in response1.status
+        assert response1.status_code == 200
+
+        # check for expected server datashape
+        response2 = empty_server.get('/datashape')
+        expected2 = str(discover({'iris': resource(iris_path)}))
+        assert response2.data.decode('utf-8') == expected2
+
+        # compute on added data
+        t = Data({'iris': resource(iris_path)})
+        expr = t.iris.petal_length.sum()
+
+        response3 = empty_server.post(
+            '/compute',
+            data=serial.dumps({'expr': to_tree(expr)}),
+            headers=mimetype(serial)
+        )
+
+        result3 = serial.loads(response3.data)['data']
+        expected3 = compute(expr, {'iris': resource(iris_path)})
+        assert result3 == expected3
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_add_data_to_server(serial):
+    with temp_server(data) as test:
+        # add data
+        initial_datashape = datashape.dshape(test.get('/datashape').data.decode('utf-8'))
+        iris_path = example('iris.csv')
+        blob = serial.dumps({'iris': iris_path})
+        response1 = test.post(
+            '/add',
+            headers=mimetype(serial),
+            data=blob,
+        )
+        assert 'OK' in response1.status
+        assert response1.status_code == 200
+
+        # check for expected server datashape
+        new_datashape = datashape.dshape(test.get('/datashape').data.decode('utf-8'))
+        data2 = data.copy()
+        data2.update({'iris': resource(iris_path)})
+        expected2 = datashape.dshape(discover(data2))
+        from pprint import pprint as pp
+        assert_dshape_equal(new_datashape, expected2)
+        assert new_datashape.measure.fields != initial_datashape.measure.fields
+
+        # compute on added data
+        t = Data({'iris': resource(iris_path)})
+        expr = t.iris.petal_length.sum()
+
+        response3 = test.post(
+            '/compute',
+            data=serial.dumps({'expr': to_tree(expr)}),
+            headers=mimetype(serial)
+        )
+
+        result3 = serial.loads(response3.data)['data']
+        expected3 = compute(expr, {'iris': resource(iris_path)})
+        assert result3 == expected3
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_cant_add_data_to_server(iris_server, serial):
+    # try adding more data to server
+    iris_path = example('iris.csv')
+    blob = serial.dumps({'iris': iris_path})
+    response1 = iris_server.post(
+        '/add',
+        headers=mimetype(serial),
+        data=blob,
+    )
+    assert response1.status_code == 422
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_bad_add_payload(empty_server, serial):
+    # try adding more data to server
+    blob = serial.dumps('This is not a mutable mapping.')
+    response1 = empty_server.post(
+        '/add',
+        headers=mimetype(serial),
+        data=blob,
+    )
+    assert response1.status_code == 422
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_odo_kwargs(test, serial):
+    expr = t.dumb
+    bad_query = {'expr': to_tree(expr)}
+
+    result = test.post(
+        '/compute',
+        headers=mimetype(serial),
+        data=serial.dumps(bad_query),
+    )
+    assert result.status_code == 500
+    assert b'return_df must be passed' in result.data
+
+    good_query = {
+        'expr': to_tree(expr),
+        'odo_kwargs': {
+            'return_df': odo(DumbResource.df, list),
+        },
+    }
+    result = test.post(
+        '/compute',
+        headers=mimetype(serial),
+        data=serial.dumps(good_query)
+    )
+    assert result.status_code == 200
+    data = serial.loads(result.data)
+    dshape = discover(DumbResource.df)
+    assert_dshape_equal(
+        datashape.dshape(data['datashape']),
+        dshape,
+    )
+    assert_frame_equal(
+        odo(data['data'], DataFrame, dshape=dshape),
+        DumbResource.df,
+    )
+
+
+@pytest.mark.parametrize('serial', all_formats)
+def test_compute_kwargs(test, serial):
+    expr = t.dumb.sort()
+    bad_query = {'expr': to_tree(expr)}
+
+    result = test.post(
+        '/compute',
+        headers=mimetype(serial),
+        data=serial.dumps(bad_query),
+    )
+    assert result.status_code == 500
+    assert b'return_df must be passed' in result.data
+
+    good_query = {
+        'expr': to_tree(expr),
+        'compute_kwargs': {
+            'return_df': odo(DumbResource.df, list),
+        },
+    }
+    result = test.post(
+        '/compute',
+        headers=mimetype(serial),
+        data=serial.dumps(good_query)
+    )
+    assert result.status_code == 200
+    data = serial.loads(result.data)
+    dshape = discover(DumbResource.df)
+    assert_dshape_equal(
+        datashape.dshape(data['datashape']),
+        dshape,
+    )
+    assert_frame_equal(
+        odo(data['data'], DataFrame, dshape=dshape),
+        DumbResource.df,
+    )
